/*****************************************************************************
Copyright (c) 2011 SmartVault, Inc.

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
******************************************************************************/
package com.smartvault.net;

import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;
import java.io.UnsupportedEncodingException;

import com.smartvault.exceptions.SmartVaultException;
import com.smartvault.proto.BasicUserInfoProto.BasicUserInfoResponse;
import com.smartvault.proto.DeleteFileProto.DeleteFileResponse;
import com.smartvault.proto.NonceProto.NonceResponse;
import com.smartvault.proto.ClientTokenProto.ClientTokenResponse;
import com.smartvault.proto.SignUpNewUserProto.SignUpNewUserResponse;
import com.smartvault.proto.NodeProto.NodeResponse;
import com.smartvault.proto.DelegationTokenProto.DelegationTokenResponse;
import com.smartvault.proto.SendEmailProto.SendEmailResponse;
import com.smartvault.proto.SendEmailProto.SendEmailRequest;
import com.smartvault.proto.PINEndpointProto.PINEndPointRequest;
import com.smartvault.proto.PINEndpointProto.PINEndPointResponse;

import com.smartvault.proto.ErrorProto.Error;
import com.smartvault.oauth.Base64;

import java.net.MalformedURLException;
import java.net.ProtocolException;
import java.util.*;

public final class ClientNetContext
{
    private static final String SLASH = "/"; //$NON-NLS-1$
    private static final String QUESTION = "?"; //$NON-NLS-1$
	private static final String AUTHORIZATION = "Authorization"; //$NON-NLS-1$
	private static final String BASIC = "BASIC "; //$NON-NLS-1$
	private static final String ACCEPT = "Accept"; //$NON-NLS-1$
	private static final String APPLICATION_X_PROTOBUF = "application/x-protobuf"; //$NON-NLS-1$
	private static final String CONTENT_TYPE = "Content-Type"; //$NON-NLS-1$
	private static final String ERROR_IS_NULL = "Error is null"; //$NON-NLS-1$
	public static final String GET = "GET"; //$NON-NLS-1$
    public static final String PUT = "PUT"; //$NON-NLS-1$
    public static final String DELETE = "DELETE"; //$NON-NLS-1$
    public static final String POST = "POST"; //$NON-NLS-1$
    public static final String MULTIPART_MIXED = "multipart/mixed"; //$NON-NLS-1$

	private boolean raiseError;
	private HttpURLConnection connection;
	
	private ClientNetContext(HttpURLConnection connection)
	{
		this.connection = connection;
		this.raiseError = true;
	}
	
	private InputStream getStream() throws IOException
	{
		InputStream stream = null;
		
		if (this.connection.getResponseCode() >= 400)
		{
			stream = this.connection.getErrorStream();
		}
		else	
		{
			stream = this.connection.getInputStream();		
		}
		
		return stream;
	}
	
	static String join(Collection<?> s, String delimiter) 
	{
		StringBuilder builder = new StringBuilder();
		Iterator iter = s.iterator();
		while (iter.hasNext()) 
		{
			builder.append(iter.next());
			if (!iter.hasNext()) 
			{
				break;                  
			}
			builder.append(delimiter);
		}
		return builder.toString();
	}

	
	private static URL cleanupURL(URL url) throws MalformedURLException, UnsupportedEncodingException
	{
		String pth = url.getPath();
		pth = pth.replace("//","/");
		
	
		String[] temp = pth.split("/");
		
		for (int i = 0; i < temp.length; ++i)
		{
			temp[i] = URLEncoder.encode(temp[i], "UTF-8").replace("+","%20");
		}
		
		pth = join(Arrays.asList(temp), "/");
		
		String query = url.getQuery();
		
	
		if (query != null) {
			pth = pth + QUESTION + query;
		}
		
		url = new URL(url.getProtocol(), url.getHost(), url.getPort(), pth);
		
	
		return url;
	}
	
	private static ClientNetContext open(URL url, String verb, String username, String password) throws MalformedURLException, IOException, ProtocolException, UnsupportedEncodingException
	{
		url = cleanupURL(url);
		HttpURLConnection connection = (HttpURLConnection)url.openConnection();		
		connection.setRequestMethod(verb);
		connection.setRequestProperty(CONTENT_TYPE, APPLICATION_X_PROTOBUF);
		connection.setRequestProperty(ACCEPT, APPLICATION_X_PROTOBUF);
		
		if (username != null && password != null)
		{
			String encodedCredential = Base64.encode( (username + ":" + password).getBytes() ); //$NON-NLS-1$
			connection.setRequestProperty(AUTHORIZATION, BASIC + encodedCredential);
		}
		
	
		return new ClientNetContext(connection);
	}
	
	public static ClientNetContext open(WebRequestInfo info, String noun, String verb, Map<String,String> m) throws Exception
	{
		String tmp = info.getUrl() + SLASH + noun;
		
		QueryString qs = new QueryString(m);
		
		URL url = new URL(tmp + QUESTION + qs.toString());				
		
		return open(url, verb, info.getUsername(), info.getPassword());
	}
	
	public static ClientNetContext open(WebRequestInfo info, String noun, String verb) throws Exception
	{
		URL url = new URL(info.getUrl() + SLASH + noun);
		return open(url, verb, info.getUsername(), info.getPassword());
	}

	public boolean getRaiseError()
	{
		return this.raiseError;
	}
	
	public void setRaiseError(boolean value)
	{
		this.raiseError = value;
	}

	public void writePayload(com.google.protobuf.GeneratedMessage payload) throws IOException
	{
		this.connection.setDoOutput(true);
		payload.writeTo(this.connection.getOutputStream());	
		this.connection.connect();		
	}
	
    public BasicUserInfoResponse readPayloadBasicUserInfoResponse() throws IOException, SmartVaultException
    {
		BasicUserInfoResponse response = BasicUserInfoResponse.parseFrom(getStream());
		testThrow(response.getError());	
		return response;
    }
    
    public NonceResponse readPayloadNonceResponse() throws IOException, SmartVaultException
    {	
		NonceResponse response = NonceResponse.parseFrom(getStream());
		testThrow(response.getError());
		return response;
    }

	public PINEndPointResponse readPayloadPinEndpointResponse() throws IOException, SmartVaultException
	{
		PINEndPointResponse response = PINEndPointResponse.parseFrom(getStream());
		testThrow(response.getError());
		return response;
	}
    
    public ClientTokenResponse readPayloadClientTokenResponse() throws IOException, SmartVaultException
    {	
		ClientTokenResponse response = ClientTokenResponse.parseFrom(getStream());
		testThrow(response.getError());	
		return response;
    }
    
    public SignUpNewUserResponse readPayloadSignUpNewUserResponse() throws IOException, SmartVaultException
    {
    	SignUpNewUserResponse response = SignUpNewUserResponse.parseFrom(getStream());
		testThrow(response.getError());	
		return response;
    }
    
    public SendEmailResponse readPayloadSendEmailResponse() throws IOException, SmartVaultException
    {
    	SendEmailResponse response = SendEmailResponse.parseFrom(getStream());
		testThrow(response.getError());	
		return response;
    }
    
    public NodeResponse readPayloadNodeResponse() throws IOException, SmartVaultException
    {
		
		InputStream str = getStream();
    
    	NodeResponse response = NodeResponse.parseFrom(str);
    	
    	
		testThrow(response.getError());	
		return response;
    }
    
    public DeleteFileResponse readDeleteFileResponse() throws IOException, SmartVaultException
    {
    	DeleteFileResponse response = DeleteFileResponse.parseFrom(getStream());
		testThrow(response.getError());	
		return response;
    }
    
    public DelegationTokenResponse readPayloadDelegationTokenResponse() throws IOException, SmartVaultException
    {
    	DelegationTokenResponse response = DelegationTokenResponse.parseFrom(getStream());
		testThrow(response.getError());	
		return response;
    }
    
    public NetAttachmentReader readAttachment() throws IOException
    {   	
    	String contentType = this.connection.getContentType();
    	String[] tmp = this.connection.getURL().getPath().split(SLASH);
    	String name = tmp[tmp.length - 1];
    	int contentLength = this.connection.getContentLength();
    	InputStream stream = getStream();
    	
    	NetAttachmentDescriptor descriptor = new NetAttachmentDescriptor(contentType, name, contentLength);
    	
    	return new NetAttachmentReader(descriptor, stream);
    }

    public void writePayload(com.google.protobuf.GeneratedMessage payload, InputStream str) throws Exception
    {
    	this.connection.setDoOutput(true);
    	
    	this.connection.setRequestProperty(CONTENT_TYPE, MULTIPART_MIXED);
    	
    	MultipartMimeWriter writer = new MultipartMimeWriter(this.connection.getOutputStream());

    	writer.write(payload);
    	writer.write(str);
    	
    	writer.complete();
    	
    	
    	this.connection.connect();
    	
    }
    
    private void testThrow(Error error) throws SmartVaultException
    {
		if (error == null)
		{	
			throw new IllegalArgumentException(ERROR_IS_NULL);
		}
    
		if (!error.getSuccess())
		{
			if (getRaiseError())
			{
				throw new SmartVaultException(error.getErrorCode(), error.getErrorText(), error.getCorrectiveUri(), error.getHelpUri(), error.getInstanceGuid(), error.getStatusCode());
			}
		}
    }
}